@@ -17,6 +17,8 @@ module System.FilePath.Glob.Base
    , optimize
 
    , liftP, tokToLower
+
+   , isLiteral
    ) where
 
 import Control.Arrow                     (first)
@@ -28,7 +30,10 @@ import Data.Char                         (isDigit, isAlpha, toLower)
 import Data.List                         (find, sortBy)
 import Data.List.NonEmpty                (toList)
 import Data.Maybe                        (fromMaybe)
+-- Monoid is re-exported from Prelude as of 4.8.0.0
+#if !MIN_VERSION_base(4,8,0)
 import Data.Monoid                       (Monoid, mappend, mempty, mconcat)
+#endif
 import Data.Semigroup                    (Semigroup, (<>), sconcat, stimes)
 import Data.String                       (IsString(fromString))
 import System.FilePath                   ( pathSeparator, extSeparator
@@ -48,7 +53,7 @@ import Text.Read (readPrec, lexP, parens, prec, Lexeme(Ident))
 data Token
    -- primitives
    = Literal !Char
-   | ExtSeparator                              --  .
+   | ExtSeparator                              --  .  optimized away to Literal
    | PathSeparator                             --  /
    | NonPathSeparator                          --  ?
    | CharRange !Bool [Either Char (Char,Char)] --  []
@@ -58,6 +63,7 @@ data Token
 
    -- after optimization only
    | LongLiteral !Int String
+   | Unmatchable  -- [/], or [.] at the beginning or after a path separator
    deriving (Eq)
 
 -- Note: CharRanges aren't converted, because this is tricky in general.
@@ -99,8 +105,7 @@ liftP f (Pattern pat) = Pattern (f pat)
 
 instance Show Token where
    show (Literal c)
-      | c `elem` "*?[<" || isExtSeparator c
-                            = ['[',c,']']
+      | c `elem` "*?[<"     = ['[',c,']']
       | otherwise           = assert (not $ isPathSeparator c) [c]
    show ExtSeparator        = [ extSeparator]
    show PathSeparator       = [pathSeparator]
@@ -116,6 +121,9 @@ instance Show Token where
    -- just put them at the end.
    --
    -- Also, [^x-] was sorted and should not become [^-x].
+   --
+   -- Also, for something like [/!^] or /[.^!] that got optimized to have just ^
+   -- and ! we need to add a dummy /.
    show (CharRange b r)     =
       let f = either (:[]) (\(x,y) -> [x,'-',y])
           (caret,exclamation,fs) =
@@ -134,10 +142,13 @@ instance Show Token where
                                  else (s',"")
        in concat [ "["
                  , if b then "" else "^"
+                 , if b && null beg && not (null caret && null exclamation) then "/" else ""
                  , beg, caret, exclamation, rest
                  , "]"
                  ]
 
+   show Unmatchable = "[.]"
+
 instance Show Pattern where
    showsPrec d p = showParen (d > 10) $
       showString "compile " . showsPrec (d+1) (decompile p)
@@ -512,7 +523,18 @@ charRange opts zs =
 
 
 optimize :: Pattern -> Pattern
-optimize = liftP (fin . go)
+optimize (Pattern pat) =
+   Pattern . fin $
+      case pat of
+         e : ts | e == ExtSeparator || e == Literal '.' ->
+            checkUnmatchable (Literal '.' :) (go ts)
+         _ ->
+            -- Handle the case where the whole pattern starts with a
+            -- now-literalized [.]. LongLiterals haven't been created yet so
+            -- checking for Literal suffices.
+            case go pat of
+                 Literal '.' : _ -> [Unmatchable]
+                 opat -> checkUnmatchable id opat
  where
    fin [] = []
 
@@ -520,8 +542,8 @@ optimize = liftP (fin . go)
    -- Has to be done here: we can't backtrack in go, but some cases might
    -- result in consecutive Literals being generated.
    -- E.g. "a[b]".
-   fin (x:y:xs) | isLiteral x && isLiteral y =
-      let (ls,rest) = span isLiteral xs
+   fin (x:y:xs) | isCharLiteral x && isCharLiteral y =
+      let (ls,rest) = span isCharLiteral xs
        in fin $ LongLiteral (length ls + 2)
                             (foldr (\(Literal a) -> (a:)) [] (x:y:ls))
                 : rest
@@ -545,48 +567,66 @@ optimize = liftP (fin . go)
    fin (x:xs) = x : fin xs
 
    go [] = []
+
+   -- Get rid of ExtSeparators, so that they can hopefully be combined into
+   -- LongLiterals later.
+   --
+   -- /.                      -> fine
+   -- . elsewhere             -> fine
+   -- /[.]                    -> Unmatchable
+   -- [.] at start of pattern -> handled outside 'go'
+   go (p@PathSeparator : ExtSeparator : xs) = p : Literal '.' : go xs
+   go (ExtSeparator : xs) = Literal '.' : go xs
+   go (p@PathSeparator : x@(CharRange _ _) : xs) =
+      p : case optimizeCharRange True x of
+             x'@(CharRange _ _) -> x' : go xs
+             Literal '.'        -> [Unmatchable]
+             x'                 -> go (x':xs)
+
    go (x@(CharRange _ _) : xs) =
-      case optimizeCharRange x of
+      case optimizeCharRange False x of
            x'@(CharRange _ _) -> x' : go xs
            x'                 -> go (x':xs)
 
-   -- <a-a> -> a
-   go (OpenRange (Just a) (Just b):xs)
-      | a == b = LongLiteral (length a) a : go xs
-
-   -- <a-b> -> [a-b]
-   -- a and b are guaranteed non-null
-   go (OpenRange (Just [a]) (Just [b]):xs)
-      | b > a = go $ CharRange True [Right (a,b)] : xs
+   -- Put [0-9] in front of <-> to allow compressing <->[0-9]<->. Handling the
+   -- [0-9] first in matching should also be faster in general.
+   go (o@(OpenRange Nothing Nothing) : d : xs) | d == anyDigit =
+      d : go (o : xs)
 
    go (x:xs) =
-      case find ($ x) compressors of
-           Just c  -> let (compressed,ys) = span c xs
-                       in if null compressed
-                             then x : go ys
-                             else go (x : ys)
+      case find ((== x) . fst) compressables of
+           Just (_, f) -> let (compressed,ys) = span (== x) xs
+                           in if null compressed
+                                 then x : go ys
+                                 else f (length compressed) ++ go (x : ys)
            Nothing -> x : go xs
 
-   compressors = [isStar, isStarSlash, isAnyNumber]
+   checkUnmatchable f ts = if Unmatchable `elem` ts then [Unmatchable] else f ts
 
-   isLiteral   (Literal _)                 = True
-   isLiteral   _                           = False
-   isStar      AnyNonPathSeparator         = True
-   isStar      _                           = False
-   isStarSlash AnyDirectory                = True
-   isStarSlash _                           = False
-   isAnyNumber (OpenRange Nothing Nothing) = True
-   isAnyNumber _                           = False
+   compressables = [ (AnyNonPathSeparator, const [])
+                   , (AnyDirectory, const [])
+                   , (OpenRange Nothing Nothing, \n -> replicate n anyDigit)
+                   ]
 
-optimizeCharRange :: Token -> Token
-optimizeCharRange (CharRange b_ rs) = fin b_ . go . sortCharRange $ rs
+   isCharLiteral (Literal _) = True
+   isCharLiteral _           = False
+
+   anyDigit = CharRange True [Right ('0', '9')]
+
+optimizeCharRange :: Bool -> Token -> Token
+optimizeCharRange precededBySlash (CharRange b rs) =
+   fin . stripUnmatchable . go . sortCharRange $ rs
  where
    -- [/] is interesting, it actually matches nothing at all
    -- [.] can be Literalized though, just don't make it into an ExtSeparator so
    --     that it doesn't match a leading dot
-   fin True [Left  c] | not (isPathSeparator c) = Literal c
-   fin True [Right r] | r == (minBound,maxBound) = NonPathSeparator
-   fin b x = CharRange b x
+   fin [Left  c] | b = if isPathSeparator c then Unmatchable else Literal c
+   fin [Right r] | b && r == (minBound,maxBound) = NonPathSeparator
+   fin x = CharRange b x
+
+   stripUnmatchable xs@(_:_:_) | b =
+      filter (\x -> (not precededBySlash || x /= Left '.') && x /= Left '/') xs
+   stripUnmatchable xs = xs
 
    go [] = []
 
@@ -629,7 +669,7 @@ optimizeCharRange (CharRange b_ rs) = fin b_ . go . sortCharRange $ rs
                    -- [a-cb-d] -> [a-d]
                    Just o  -> go$ Right o : ys
                    Nothing -> x : go xs
-optimizeCharRange _ = error "Glob.optimizeCharRange :: internal error"
+optimizeCharRange _ _ = error "Glob.optimizeCharRange :: internal error"
 
 sortCharRange :: [Either Char (Char,Char)] -> [Either Char (Char,Char)]
 sortCharRange = sortBy cmp
@@ -638,3 +678,13 @@ sortCharRange = sortBy cmp
    cmp (Left   a)    (Right (b,_)) = compare a b
    cmp (Right (a,_)) (Left   b)    = compare a b
    cmp (Right (a,_)) (Right (b,_)) = compare a b
+
+-- |Returns `True` iff the given `Pattern` is a literal file path, i.e. it has
+-- no wildcards, character ranges, etc.
+isLiteral :: Pattern -> Bool
+isLiteral = all lit . unPattern
+ where
+   lit (Literal _) = True
+   lit (LongLiteral _ _) = True
+   lit PathSeparator = True
+   lit _ = False
